from pathlib import Path
from typing import Union

from .base import ReaderBase, convert_image


# NOTE: This is a placeholder for the actual implementation
#      of the CuCIM reader. It's not tested and will not work.
# TODO: Implement the CuCIM reader on GPU-available machine
class CuCIMReader(ReaderBase):
    """
    Use CuCIM to interface with image files.

    See `CuCIM <https://github.com/rapidsai/cucim>`_ for more information.

    Parameters
    ----------
    file : str or Path
        Path to image file on disk

    """

    def __init__(
        self,
        file: Union[Path, str],
        **kwargs,
    ):
        self.file = str(file)
        self.create_reader()
        self.set_properties(self._reader.properties)

    def get_region(
        self,
        x,
        y,
        width,
        height,
        level: int = 0,
        **kwargs,
    ):
        level = self.translate_level(level)
        img = self.reader.read_region(
            (x, y),
            (int(width), int(height)),
            level=level,
        )
        return convert_image(img)

    def get_level(self, level):
        level = self.translate_level(level)
        img = self.reader.read_region(
            (0, 0), level, self.reader.level_dimensions[level]
        )
        return convert_image(img)

    def detach_reader(self):
        if self._reader is not None:
            self._reader.close()
            self._reader = None

    def create_reader(self):
        from cucim import CuImage

        self._reader = CuImage(self.file)

    @property
    def reader(self):
        if self._reader is None:
            self.create_reader()
        return self._reader



from __future__ import annotations

__all__ = [
    "add_tissues",
    "add_shapes",
    "add_tiles",
    "add_features",
    "add_agg_features",
    "add_table",
    "update_shapes_data",
    "subset_tiles",
]

from typing import Sequence, Mapping

import numpy as np
import pandas as pd
import geopandas as gpd
from anndata import AnnData
from shapely import box, Polygon
from spatialdata.models import ShapesModel, TableModel

from .._model import WSIData, TileSpec


def add_tissues(
    wsidata: WSIData, key: str, tissues: Sequence[Polygon], ids=None, **kws
):
    """Add tissue contours to the SpatialData.

    - :bdg-primary:`SpatialData Slot`: :bdg-danger:`shapes`
    - :bdg-info:`Table Columns`: :bdg-black:`tissue_id`, :bdg-black:`geometry`

    Parameters
    ----------
    wsidata : WSIData
        The WSIData object.
    key : str
        The key of the tissue contours.
    tissues : array of :class:`Polygon <shapely.Polygon>`
        A sequence of shapely Polygon objects.
    ids : Sequence[int], optional
        The tissue ids.


    """
    if ids is None:
        ids = np.arange(len(tissues))
    gdf = gpd.GeoDataFrame(
        data={
            "tissue_id": ids,
            "geometry": tissues,
        }
    )
    cs = ShapesModel.parse(gdf, **kws)
    wsidata.shapes[key] = cs


def add_tiles(wsidata, key, xys, tile_spec: TileSpec, tissue_ids, **kws):
    # Tiles should be stored as polygon
    # This allows easy query of which cells in which tiles
    w, h = tile_spec.base_width, tile_spec.base_height
    gdf = gpd.GeoDataFrame(
        data={
            "tile_id": np.arange(len(xys)),
            "x": xys[:, 0],
            "y": xys[:, 1],
            "tissue_id": tissue_ids,
            "geometry": [box(x, y, x + w, y + h) for (x, y) in xys],
        }
    )
    cs = ShapesModel.parse(gdf, **kws)
    wsidata.shapes[key] = cs

    if wsidata.TILE_SPEC_KEY in wsidata.attrs:
        spec_data = wsidata.attrs[wsidata.TILE_SPEC_KEY]
        spec_data[key] = tile_spec.to_dict()
    else:
        spec_data = {key: tile_spec.to_dict()}
        wsidata.attrs[wsidata.TILE_SPEC_KEY] = spec_data


def subset_tiles(wsidata, key, indices, new_key=None):
    shapes = wsidata.shapes[key]
    shapes = shapes.iloc[indices]
    if new_key is None:
        new_key = key
    else:
        spec_data = wsidata.attrs[wsidata.TILE_SPEC_KEY]
        spec_data[new_key] = spec_data[key]
    wsidata.shapes[new_key] = shapes


def add_features(wsidata, key, tile_key, features, **kws):
    tile_tb = wsidata[tile_key]
    if "tile_id" in tile_tb.columns:
        tile_id = tile_tb["tile_id"].values
    else:
        tile_id = np.arange(len(features))
    library_id = pd.Categorical([tile_key])

    if "obs" in kws:
        obs = kws["obs"]
        obs["tile_id"] = tile_id
        obs["library_id"] = library_id
    else:
        obs = pd.DataFrame(
            {"tile_id": tile_id, "library_id": library_id},
            index=tile_id.astype(str),
        )
        kws["obs"] = obs
    adata = AnnData(X=features, **kws)

    model_kws = dict(region=tile_key, region_key="library_id", instance_key="tile_id")
    feature_table = TableModel.parse(adata, **model_kws)
    wsidata.tables[key] = feature_table


def add_agg_features(wsidata, key, agg_key, agg_features, by_key=None, by_data=None):
    feature_table = wsidata.tables[key]
    feature_table.varm[agg_key] = agg_features

    agg_ops = feature_table.uns.get("agg_ops", {})
    agg_ops[agg_key] = by_key
    feature_table.uns["agg_ops"] = agg_ops
    if by_data is not None:
        feature_table.obs[by_key] = by_data


# The following might be unnecessary? Should we remove it?
def add_shapes(wsidata, key, shapes, **kws):
    cs = ShapesModel.parse(shapes, **kws)
    wsidata.shapes[key] = cs


def add_table(wsidata, key, table, **kws):
    table = TableModel.parse(table, **kws)
    wsidata.tables[key] = table


def update_shapes_data(wsidata, key: str, data: Mapping | pd.DataFrame):
    shapes = wsidata.shapes[key]
    if isinstance(data, pd.DataFrame):
        data = data.to_dict(orient="series")
    for k, v in data.items():
        shapes[k] = v



from ._wsi import open_wsi, agg_wsi
from ._elems import (
    add_tissues,
    add_shapes,
    add_tiles,
    add_features,
    add_agg_features,
    add_table,
    update_shapes_data,
    subset_tiles,
)



from torch.utils.data import Dataset

from .._model import WSIData


def graph_data(wsi: WSIData, key: str = "graph", target_key: str = None):
    tables = wsi.tables[key]
    X = tables.X
    tables = tables.obs
    targets = None
    if target_key in tables:
        targets = tables[target_key].to_numpy()
    return X, targets



import torch


def transform(tensor):
    # Convert to float32 and scale (assuming input is uint8 or another dtype)
    if not isinstance(tensor, torch.Tensor):
        tensor = torch.tensor(tensor)
    tensor = tensor.to(torch.float32)
    # Scale if needed (assuming input is originally in [0,1] or [0,255])
    if tensor.max() <= 1:  # If already normalized in [0,1], scale to [0,255]
        tensor = tensor * 255
    return tensor


class ColorNormalizer(torch.nn.Module):
    def __init__(self, method="macenko"):
        super().__init__()

        try:
            import torchstain.torch.normalizers as norm
            import torchstain.torch.augmentors as aug
        except (ImportError, ModuleNotFoundError):
            raise ImportError("To use color normalization, please install torchstain.")

        self.method = method
        if method == "macenko":
            normalizer = norm.TorchMacenkoNormalizer()
        elif method == "reinhard":
            normalizer = norm.TorchReinhardNormalizer()
            normalizer.target_means = torch.tensor([72.909996, 20.8268, -4.9465137])
            normalizer.target_stds = torch.tensor([18.560713, 14.889295, 5.6756697])
        elif method == "reinhard_modified":
            normalizer = norm.TorchReinhardNormalizer(method="modified")
            normalizer.target_means = torch.tensor([72.909996, 20.8268, -4.9465137])
            normalizer.target_stds = torch.tensor([18.560713, 14.889295, 5.6756697])
        elif method == "multi_macenko":
            normalizer = norm.TorchMultiMacenkoNormalizer()
        elif method == "macenko_aug":
            normalizer = aug.TorchMacenkoAugmentor()
        else:
            raise NotImplementedError(f"Requested method '{method}' not implemented")
        self.normalizer = normalizer

    def __repr__(self):
        return f"ColorNormalizer(method='{self.method}')"

    def forward(self, img):
        t_img = transform(img)
        if self.method == "macenko":
            norm, _, _ = self.normalizer.normalize(I=t_img)
        else:
            norm = self.normalizer.normalize(I=t_img)
        return norm



from pathlib import Path
from typing import Union

from .base import ReaderBase, convert_image


class OpenSlideReader(ReaderBase):
    """
    Use OpenSlide to interface with image files.

    Depends on `openslide-python <https://openslide.org/api/python/>`_
    which wraps the `openslide <https://openslide.org/>`_ C library.

    Parameters
    ----------
    file : str or Path
        Path to image file on disk

    """

    name = "openslide"

    def __init__(
        self,
        file: Union[Path, str],
        **kwargs,
    ):
        self.file = str(file)
        self.create_reader()
        self.set_properties(self._reader.properties)

    def get_region(
        self,
        x,
        y,
        width,
        height,
        level: int = 0,
        **kwargs,
    ):
        level = self.translate_level(level)
        # All types are coerced to native Python types
        img = self.reader.read_region(
            (int(x), int(y)), int(level), (int(width), int(height))
        )
        return convert_image(img)

    def get_thumbnail(self, size, **kwargs):
        sx, sy = self.properties.shape
        if size > sx or size > sy:
            raise ValueError("Requested thumbnail size is larger than the image")
        # The size is only the maximum size
        if sx > sy:
            size = (size, int(size * sy / sx))
        else:
            size = (int(size * sx / sy), size)

        img = self.reader.get_thumbnail(size)
        return convert_image(img)

    def detach_reader(self):
        if self._reader is not None:
            try:
                self._reader.close()
                self._reader = None
            # There is a chance that the pointer
            # to C-library is already collected
            except TypeError:
                pass

    def create_reader(self):
        from openslide import OpenSlide

        self._reader = OpenSlide(self.file)



from __future__ import annotations

import base64
import io
import json
import warnings
from dataclasses import dataclass, asdict
from functools import cached_property
from numbers import Integral, Number
from pathlib import Path
from typing import Literal, Generator, Sequence, TYPE_CHECKING

import numpy as np
from PIL.Image import Image, fromarray
from anndata import AnnData
from ome_zarr.io import parse_url
from spatialdata import SpatialData
from spatialdata.models import SpatialElement

from .._utils import find_stack_level
from ..accessors import FetchAccessor, IterAccessor, DatasetAccessor
from ..reader import ReaderBase, SlideProperties

if TYPE_CHECKING:
    from typing import Self


class WSIData(SpatialData):
    """
    A container class combining :class:`SpatialData <spatialdata.SpatialData>`
    and a whole slide image reader.

    .. note::
       Use the :func:`open_wsi` function to create a WSIData object.

    By default, the whole slide image is not attached to the SpatialData.
    A thumbnail version of the whole slide image is attached for visualization purpose.

    .. list-table::
       :header-rows: 1

       * - **Content**
         - **Default key**
         - **Slot**
         - **Type**

       * - Whole slide image
         - :bdg-info:`wsi_thumbnail`
         - :bdg-danger:`images`
         - :class:`DataArray <xarray.DataArray>` (c, y, x) format

       * - Slide Properties
         - :bdg-info:`slide_properties`
         - :bdg-danger:`attrs`
         - :class:`SlideProperties <wsidata.SlideProperties>`

       * - Tissue contours
         - :bdg-info:`tissues`
         - :bdg-danger:`shapes`
         - :class:`GeoDataFrame <geopandas.GeoDataFrame>`

       * - Tile locations
         - :bdg-info:`tiles`
         - :bdg-danger:`shapes`
         - :class:`GeoDataFrame <geopandas.GeoDataFrame>`

       * - Tile specifications
         - :bdg-info:`tile_spec`
         - :bdg-danger:`attrs`
         - :class:`TileSpec <wsidata.TileSpec>`

       * - Features
         - :bdg-info:`{feature_key}_{tile_key}`
         - :bdg-danger:`tables`
         - :class:`AnnData <anndata.AnnData>`


    You can interact with WSIData using the following accessors:

    - :class:`fetch <wsidata.FetchAccessor>`: Access data from the WSIData object.
    - :class:`iter <wsidata.IterAccessor>`: Iterate over data in the WSIData object.
    - :class:`ds <wsidata.DatasetAccessor>`: Create deep learning datasets from the WSIData object.
    - To implement your own accessors, use :func:`register_wsidata_accessor <wsidata.register_wsidata_accessor>`.

    For analysis purpose, you can override two slide properties:

    - **microns per pixel (mpp)**: Using the :meth:`set_mpp` method.
    - **bounds**: Using the :meth:`set_bounds` method.

    Other Parameters
    ----------------
    reader : :class:`ReaderBase <wsidata.reader.ReaderBase>`
        A reader object that can interface with the whole slide image file.
    slide_properties_source : {'slide', 'sdata'}, default: 'sdata'
        The source of the slide properties.

        - "slide": load from the reader object.
        - "sdata": load from the SpatialData object.

    Attributes
    ----------
    properties : :class:`SlideProperties <wsidata.SlideProperties>`
        The properties of the whole slide image.
    reader : :class:`ReaderBase <wsidata.reader.ReaderBase>`
        The reader object for interfacing with the whole slide image.
    wsi_store :
        The store path for the whole slide image.

    """

    TILE_SPEC_KEY = "tile_spec"
    SLIDE_PROPERTIES_KEY = "slide_properties"

    @classmethod
    def from_spatialdata(cls, sdata, reader=None, **kws):
        d = cls(
            images=sdata.images,
            labels=sdata.labels,
            shapes=sdata.shapes,
            tables=sdata.tables,
            points=sdata.points,
            attrs=sdata.attrs,
            reader=reader,
            **kws,
        )
        d.path = sdata.path
        return d

    def __init__(
        self,
        images=None,
        labels=None,
        shapes=None,
        tables=None,
        points=None,
        attrs=None,
        reader: ReaderBase = None,
        slide_properties_source: Literal["slide", "sdata"] = "sdata",
    ):
        self._reader = reader
        self._wsi_store = None
        self.slide_properties_source = slide_properties_source
        self._exclude_elements = set()
        super().__init__(
            images=images,
            labels=labels,
            shapes=shapes,
            tables=tables,
            points=points,
            attrs=attrs,
        )

        if self.SLIDE_PROPERTIES_KEY not in self:
            self.attrs[self.SLIDE_PROPERTIES_KEY] = reader.properties.to_dict()
        else:
            # Try to load the slide properties from the spatial data
            if slide_properties_source == "slide":
                reader_properties = self.attrs[self.SLIDE_PROPERTIES_KEY]
                if reader_properties != reader.properties.to_dict():
                    # Update the reader properties
                    reader.properties.from_mapping(reader_properties)
                    warnings.warn(
                        "Slide properties in the spatial data is different from the reader properties.",
                        UserWarning,
                        stacklevel=find_stack_level(),
                    )

    def __repr_texts(self):
        H, W = self.properties.shape
        dimension_text = f"{H}×{W} (h×w)"
        n_level = self.properties.n_level
        pyramid_text = f"{n_level} {'Pyramid' if n_level == 1 else 'Pyramids'}"
        mpp_text = "Unknown"
        if self.properties.mpp is not None:
            mpp_text = f"{self.properties.mpp:.2f} MPP"
        if self.properties.magnification is not None:
            mpp_text += f" ({int(self.properties.magnification)}X)"
        return dimension_text, pyramid_text, mpp_text

    def __repr__(self):
        dimension_text, pyramid_text, mpp_text = self.__repr_texts()

        return (
            f"WSI: {self.reader.file}\n"
            f"Reader: {self.reader.name}\n"
            f"Dimensions: {dimension_text}, {pyramid_text}\n"
            f"Pixel physical size: {self.properties.mpp} MPP\n"
            f"{super().__repr__()}"
        )

    def _repr_html_(self):
        spatialdata_repr = super().__repr__()
        dimension_text, pyramid_text, mpp_text = self.__repr_texts()

        return f"""
                <div style="display: flex; align-items: center; gap: 10px;">
                    <img src="data:image/png;base64,{self._html_thumbnail}" 
                    style="border: 1px solid #ddd; border-radius: 8px; 
                    max-width: 300px; max-height: 100%; object-fit: contain;">
                    <div>
                        <b>WSI:</b> {self.reader.file}<br>
                        <b>Reader:</b> {self.reader.name}<br>
                        <b>Dimensions:</b> {dimension_text}, {pyramid_text}<br>
                        <b>Pixel physical size:</b> {mpp_text}<br>
                        <pre style="padding: 8px; border-radius: 4px; font-size: 8pt">{spatialdata_repr}</pre>
                    </div>
                </div>"""

    def set_exclude_elements(self, elements):
        """Set the elements to be excluded from serialize to the WSIData object on disk."""
        self._exclude_elements.update(elements)

    def set_wsi_store(self, store: str | Path):
        """Set the on disk path for the WSIData."""
        self._wsi_store = Path(store)

    def _gen_elements(
        self, include_table: bool = False
    ) -> Generator[tuple[str, str, SpatialElement | AnnData], None, None]:
        for i in super()._gen_elements(include_table):
            if i[1] not in self._exclude_elements:
                yield i

    def close(self):
        """Close the reader object."""
        self.reader.detach_reader()

    @property
    def reader(self) -> ReaderBase:
        """The reader object for interfacing with the whole slide image."""
        return self._reader

    @property
    def properties(self) -> SlideProperties:
        """The properties of the whole slide image. See :class:`SlideProperties <wsidata.SlideProperties>`."""
        return self.reader.properties

    @property
    def wsi_store(self) -> str | Path:
        """The zarr store path for the associated data of the whole slide image."""
        return self._wsi_store

    @property
    def thumbnail(self) -> Image:
        """The thumbnail of the whole slide image."""
        return self.get_thumbnail(size=500, as_array=False)

    @cached_property
    def _html_thumbnail(self) -> str:
        buffer = io.BytesIO()
        image = self.get_thumbnail(size=250, as_array=False)
        image.save(buffer, format="PNG")
        return base64.b64encode(buffer.getvalue()).decode("utf-8")

    @property
    def name(self) -> str:
        """The file name of the whole slide image."""
        return Path(self.reader.file).name

    def tile_spec(self, key: str) -> TileSpec | None:
        """
        Get the :class:`TileSpec` for a collection of tiles.

        Parameters
        ----------
        key : str
            The key of the tiles.

        """
        if self.TILE_SPEC_KEY in self.attrs:
            spec = self.attrs[self.TILE_SPEC_KEY].get(key)
            if spec is not None:
                return TileSpec(**spec)

    def set_mpp(self, mpp):
        """Set the microns per pixel (mpp) of the whole slide image.
        This will override the recorded mpp in the slide properties.
        Could be useful when the mpp is not recorded in the image file.

        Parameters
        ----------
        mpp : float
            The microns per pixel.

        """
        self.properties.mpp = mpp
        self.attrs[self.SLIDE_PROPERTIES_KEY]["mpp"] = mpp

    def set_bounds(self, bounds: tuple[int, int, int, int]):
        """Set the bounds of the whole slide image.

        Parameters
        ----------
        bounds : tuple[int, int, int, int]
            The bounds of the whole slide image in the format [x, y, width, height].

        """
        self.properties.bounds = bounds
        self.tables[self.SLIDE_PROPERTIES_KEY].uns["bounds"] = bounds

    def read_region(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        level: int = 0,
        **kwargs,
    ) -> np.ndarray[np.uint8]:
        """Read a region from the whole slide image.

        Parameters
        ----------
        x : int
            The x-coordinate at level 0.
        y : int
            The y-coordinate at level 0.
        width : int
            The width of the region in pixels.
        height : int
            The height of the region in pixels.
        level : int, default: 0
            The pyramid level.

        """
        if self.reader is None:
            raise ValueError("The reader object is not attached.")
        return self.reader.get_region(x, y, width, height, level=level, **kwargs)

    def get_thumbnail(self, size=500, as_array=False) -> np.ndarray[np.uint8] | Image:
        """Get the thumbnail of the whole slide image.

        Parameters
        ----------
        as_array : bool, default: False
            Return as numpy array.

        """
        img = self.reader.get_thumbnail(size=size)
        if as_array:
            return img
        else:
            return fromarray(img)

    def write(
        self,
        file_path=None,
        overwrite: bool = True,
        consolidate_metadata: bool = True,
        format=None,
    ):
        if file_path is not None:
            file_path = Path(file_path)
            if self.path is None:
                self.path = file_path
        else:
            if self._wsi_store is None and self.path is None:
                raise ValueError(
                    "The store path for the WSIData is not set. "
                    "Please set the store path before saving."
                )
            file_path = self._wsi_store
        super().write(
            file_path=file_path,
            overwrite=overwrite,
            consolidate_metadata=consolidate_metadata,
            format=format,
        )

    def to_spatialdata(self) -> SpatialData:
        """
        Convert the WSIData object to a SpatialData object.

        .. note::
            This is not a deep copy operation.
            Any changes to the returned SpatialData will affect the original WSIData.

        """
        return SpatialData(
            images=self.images,  # noqa
            labels=self.labels,  # noqa
            shapes=self.shapes,  # noqa
            tables=self.tables,
            points=self.points,  # noqa
            attrs=self.attrs,
        )

    def _validate_can_safely_write_to_path(
        self,
        file_path: str | Path,
        overwrite: bool = False,
        saving_an_element: bool = False,
    ) -> None:
        if isinstance(file_path, str):
            file_path = Path(file_path)

        if not isinstance(file_path, Path):
            raise ValueError(
                f"file_path must be a string or a Path object, type(file_path) = {type(file_path)}."
            )

        if Path(file_path).exists():
            if parse_url(file_path, mode="r") is None:
                raise ValueError(
                    "The target file path specified already exists, and it has been detected to not be a Zarr store. "
                    "Overwriting non-Zarr stores is not supported to prevent accidental data loss."
                )
            if not overwrite:
                raise ValueError(
                    "The Zarr store already exists. Use `overwrite=True` to try overwriting the store."
                    "Please note that only Zarr stores not currently in used by the current SpatialData object can be "
                    "overwritten."
                )
        # Skip the workaround for now

    def _check_feature_key(self, feature_key, tile_key=None):
        msg = f"{feature_key} doesn't exist"
        if feature_key in self:
            return feature_key
        else:
            if tile_key is not None:
                feature_key = f"{feature_key}_{tile_key}"
                if feature_key in self:
                    return feature_key
                msg = f"Neither {feature_key} or {feature_key}_{tile_key} exist"

        raise KeyError(msg)

    # Define default accessors as property,
    # this will enable auto-completion in IDE
    # For extension, user can define their own accessors
    # using register_wsidata_accessor
    @cached_property
    def fetch(self):
        return FetchAccessor(self)

    @cached_property
    def iter(self):
        return IterAccessor(self)

    @cached_property
    def ds(self):
        return DatasetAccessor(self)


@dataclass
class TileSpec:
    """Data class for storing tile specifications.

    To enable efficient tils generation, there are 3 levels of tile size:

    1. The destined tile size and level requested by the user
    2. The tile size and level used by the image reader to optimize the performance
    3. The actual tile size at the level 0

    This enables user to request tile size and level that are not exist in the image pyramids.
    For example, if our slide is mpp=0.5 (20X) with pyramids of mpp=[0.5, 2.0, 4.0],
    and user request mpp=1.0 (10X) with tile size 512x512. There is no direct level to read from,
    we need to read from mpp=0.5 with tile size of 1024x1024 and downsample to 512x512.


    Parameters
    ----------
    height : int
        The height of the tile.
    width : int
        The width of the tile.
    stride_height : int
        The height of the stride.
    stride_width : int
        The width of the stride.
    mpp : float, default: None
        The requested microns per pixel of tiles.
    ops_level : int, default: 0
        The level of the tiling operation.
    ops_downsample : float, default: 1
        The downsample factor to transform the operation level
        tiles to the requested level.
    tissue_name : str, optional
        The name of the tissue.

    Attributes
    ----------
    ops_{height, width} : int
        The height/width of the tile when retrieving images.
    ops_stride_{height, width}: int
        The height/width of the stride when retrieving images.
    base_{height, width} : int
        The height/width of the tile at the level 0.
    base_stride_{height, width} : int
        The height/width of the stride at the level 0.

    """

    height: int
    width: int
    stride_height: int
    stride_width: int

    mpp: float | None = None

    ops_level: int = 0  # level of the tiling operation
    ops_downsample: float = 1  # downsample to requested level

    base_level: int = 0  # level of the base tile, always 0
    base_downsample: float = 1.0  # downsample to requested level

    tissue_name: str | None = None

    def __post_init__(self):
        # coerce attributes to correct type
        self.height = int(self.height)
        self.width = int(self.width)
        self.stride_height = int(self.stride_height)
        self.stride_width = int(self.stride_width)
        self.ops_level = int(self.ops_level)
        self.ops_downsample = float(self.ops_downsample)
        self.base_level = int(self.base_level)
        self.base_downsample = float(self.base_downsample)

    def __repr__(self):
        return (
            f"Tile at {self.mpp} mpp, {self.height}×{self.width} (h×w)\n"
            f"Stride: {self.stride_height}×{self.stride_width} ({self.overlap_y}×{self.overlap_x} overlap)\n"
            f"Operation size: {self.ops_height}×{self.ops_width}, level={self.ops_level}\n"
            f"Base size: {self.base_height}×{self.base_width}, level={self.base_level}\n"
            f"Target tissue: '{self.tissue_name}'"
        )

    def _repr_html_(self):
        BORDER_WIDTH = 2
        scale = 100 / max(self.width, self.height)
        scaled_w = np.floor(self.width * scale) - BORDER_WIDTH
        scaled_h = np.floor(self.height * scale) - BORDER_WIDTH
        scaled_stride_w = np.floor(
            self.stride_width * scale
        )  # add 2 to account for the border
        scaled_stride_h = np.floor(
            self.stride_height * scale
        )  # add 2 to account for the border

        tile_style = (
            f"width: {int(scaled_w)}px; "
            f"height: {int(scaled_h)}px; "
            f"position: absolute; "
            f"box-sizing: content-box;"  # make sure the border is included in the size
            f"border: {BORDER_WIDTH}px solid #000; "
        )
        container_style = (
            f"width: {int(scaled_w + scaled_stride_w) + 2 * BORDER_WIDTH}px; "
            f"height: {int(scaled_h + scaled_stride_h) + 2 * BORDER_WIDTH}px; "
            f"position: relative;"
            f"border: 0;"
        )
        html = f"""
        <div style="display: flex; align-item: center; gap: 10px;">
            <div style='{container_style}'>
                <div style='{tile_style}; top: 0px; left: 0px; border-color: #C68FE6'>
                    <p style='padding-left: 3pt'>Tile 1</p>
                </div>
                <div style='{tile_style}; top: 0px; left: {scaled_stride_w}px; border-color: #F6DC43'>
                    <p style='padding-left: 3pt'>Tile 2</p>
                </div>
                <div style='{tile_style}; top: {scaled_stride_w}px; left: 0px; border-color: #6A9C89'>
                    <p style='padding-left: 3pt'>Tile 3</p>
                </div>
            </div>
            <div>
                <b>Tile at</b>: {self.mpp} mpp<br>
                <b>Tile size</b>: {self.height}×{self.width} (h×w)<br>
                <b>Stride</b>: {self.stride_height}×{self.stride_width} ({self.overlap_y}×{self.overlap_x} overlap)<br>
                <b>Operation size</b>: {self.ops_height}×{self.ops_width}, level={self.ops_level}<br>
                <b>Base size</b>: {self.base_height}×{self.base_width}, level={self.base_level}<br>
                <b>Target tissue</b>: '{self.tissue_name}'
            </div>
        </div>
        """
        return html

    @classmethod
    def from_wsidata(
        cls,
        wsi: WSIData,
        tile_px: int | (int, int),
        stride_px: int | (int, int) = None,
        overlap: float | (float, float) = None,
        mpp=None,
        ops_level=None,
        slide_mpp=None,
        tissue_name=None,
    ) -> Self:
        """Create a TileSpec from a WSIData object.

        To tile from the whole slide image, the user needs to specify the tile size and stride size.
        mpp only need to be specified if the user wants to make sure
        the tile size is harmonized across different slides.

        If ops_level is not specified, the optimal level will be calculated based on the requested mpp
        to maximize the performance.

        """
        # Check if the input params are valid
        tile_w, tile_h, stride_w, stride_h = _preprocess_tile_params(
            tile_px, stride_px, overlap
        )

        # If user does not override slide mpp, use the recorded slide mpp
        if slide_mpp is None:
            slide_mpp = wsi.properties.mpp

        # If user does not specify mpp, default to ops level is 0
        if mpp is None:
            mpp = wsi.properties.mpp
            if ops_level is not None:
                raise ValueError("Please specify mpp if ops_level is specified.")
            ops_level = 0
            ops_downsample = 1
        # Or if user specify mpp, but the slide mpp is not available, use the default level 0
        elif slide_mpp is None:
            if ops_level is None:
                ops_level = 0
                ops_downsample = 1
            else:
                ops_downsample = wsi.properties.level_downsample[ops_level]

            warning_text = f"Slide mpp is not available, using level {ops_level}."
            # If user specify the mpp at the same time but slide mpp is not available
            # we will inform the user that the mpp will be ignored
            if mpp is not None:
                warning_text += f" Requested mpp={mpp} will be ignored."
            warnings.warn(warning_text, stacklevel=find_stack_level())
        else:
            # If user didn't specify ops level but mpp, an optimized ops level will be calculated
            wsi_downsample = np.asarray(wsi.properties.level_downsample)
            req_downsample = mpp / slide_mpp
            if req_downsample < 1:
                raise ValueError(
                    f"Requested mpp={mpp} is smaller than the slide mpp={slide_mpp}. "
                    f"Up-sampling is not supported."
                )

            gap = wsi_downsample - req_downsample
            gap[gap > 0] = np.inf

            suggest_level = np.argmin(np.abs(gap))
            suggest_downsample = mpp / (slide_mpp * wsi_downsample[suggest_level])
            if ops_level is None:
                # Apply the optimized level
                ops_level = suggest_level
                ops_downsample = suggest_downsample
            else:
                # check if user specified ops level is valid
                ops_level = wsi.reader.translate_level(ops_level)
                ops_downsample = mpp / (
                    slide_mpp * wsi.properties.level_downsample[ops_level]
                )
                if ops_downsample < 1:
                    if suggest_level == 0:
                        consider_text = "Consider using level 0."
                    else:
                        consider_text = f"Consider using ops_level<={suggest_level}."
                    raise ValueError(
                        f"Requested tile size at mpp={mpp} "
                        f"on ops_level={ops_level} will require up-sampling. "
                        + consider_text
                    )

        base_downsample = ops_downsample * wsi.properties.level_downsample[ops_level]

        return cls(
            height=tile_h,
            width=tile_w,
            stride_height=stride_h,
            stride_width=stride_w,
            mpp=mpp,
            ops_level=ops_level,
            ops_downsample=ops_downsample,
            base_downsample=base_downsample,
            tissue_name=tissue_name,
        )

    def to_dict(self):
        """Convert the TileSpec to a dictionary."""
        return asdict(self)

    def to_json(self):
        """Convert the TileSpec to a JSON string."""
        return json.dumps(asdict(self))

    @cached_property
    def is_overlap_x(self) -> bool:
        """Check if the tiles are overlapped along the x-axis."""
        return self.stride_width < self.width

    @cached_property
    def is_overlap_y(self) -> bool:
        """Check if the tiles are overlapped along the y-axis."""
        return self.stride_height < self.height

    @cached_property
    def is_overlap(self) -> bool:
        """Check if the tiles are overlapped."""
        return self.is_overlap_x or self.is_overlap_y

    @cached_property
    def overlap_x(self) -> int:
        """The overlap pixel size along the x-axis."""
        return self.width - self.stride_width

    @cached_property
    def overlap_y(self) -> int:
        """The overlap pixel size along the y-axis."""
        return self.height - self.stride_height

    @cached_property
    def ops_height(self) -> int:
        """The height of the tile when retrieving images."""
        return int(self.height * self.ops_downsample)

    @cached_property
    def ops_width(self) -> int:
        """The width of the tile when retrieving images."""
        return int(self.width * self.ops_downsample)

    @cached_property
    def ops_stride_height(self) -> int:
        """The height of the stride when retrieving images."""
        return int(self.stride_height * self.ops_downsample)

    @cached_property
    def ops_stride_width(self) -> int:
        """The width of the stride when retrieving images."""
        return int(self.stride_width * self.ops_downsample)

    @cached_property
    def base_height(self) -> int:
        """The height of the tile at the level 0."""
        return int(self.height * self.base_downsample)

    @cached_property
    def base_width(self) -> int:
        """The width of the tile at the level 0."""
        return int(self.width * self.base_downsample)

    @cached_property
    def base_stride_height(self) -> int:
        """The height of the stride at the level 0."""
        return int(self.stride_height * self.base_downsample)

    @cached_property
    def base_stride_width(self) -> int:
        """The width of the stride at the level 0."""
        return int(self.stride_width * self.base_downsample)


def _preprocess_tile_params(
    tile_px: int | (int, int),
    stride_px: int | (int, int) = None,
    overlap: Number | (Number, Number) = 0.5,
):
    if isinstance(tile_px, Integral):
        tile_w, tile_h = (tile_px, tile_px)
    elif isinstance(tile_px, Sequence):
        tile_w, tile_h = (tile_px[0], tile_px[1])
    else:
        raise TypeError("tile_px must be an integer or a tuple of two integers.")

    if tile_w <= 0 or tile_h <= 0:
        raise ValueError("tile_px must be positive.")

    if stride_px is not None and overlap is not None:
        raise ValueError("Cannot specify both stride_px and overlap.")

    stride_w, stride_h = tile_w, tile_h
    if stride_px is not None:
        if isinstance(stride_px, Integral):
            stride_w, stride_h = (stride_px, stride_px)
        elif isinstance(stride_px, Sequence):
            stride_w, stride_h = (stride_px[0], stride_px[1])
        else:
            raise TypeError("stride_px must be an integer or a tuple of two integers.")
        if stride_w <= 0 or stride_h <= 0:
            raise ValueError("stride_px must be positive.")

    if overlap is not None:
        if isinstance(overlap, Number):
            overlap_w, overlap_h = (overlap, overlap)
        elif isinstance(overlap, Sequence):
            overlap_w, overlap_h = (overlap[0], overlap[1])
        else:
            raise TypeError("overlap must be a number or a tuple of two numbers.")

        if overlap_w < 0 or overlap_h < 0:
            raise ValueError("overlap must be non-negative.")
        # calculate stride px from overlap

        if overlap_w < 1:
            stride_w = int(tile_w * (1 - overlap_w))
        else:
            stride_w = int(tile_w - overlap_w)

        if overlap_h < 1:
            stride_h = int(tile_h * (1 - overlap_h))
        else:
            stride_h = int(tile_h - overlap_h)

    return tile_w, tile_h, stride_w, stride_h



__all__ = ["to_datatree"]


# Adapted from https://github.com/gustaveroussy/sopa/blob/master/sopa/io/reader/wsi.py
from ctypes import ArgumentError
from dataclasses import asdict
from functools import singledispatch
from typing import Dict

import numpy as np
from zarr.storage import (
    _path_to_prefix,
    attrs_key,
    init_array,
    init_group,
    Store,
    KVStore,
)
from zarr.util import json_dumps, normalize_storage_path, normalize_shape

from ..reader import TiffSlideReader
from ..reader.base import ReaderBase

from xarray import open_zarr, DataArray, DataTree, Dataset
from spatialdata.models import Image2DModel
from spatialdata.transformations import Identity, Scale


def init_attrs(store, attrs, path: str = None):
    path = normalize_storage_path(path)
    path = _path_to_prefix(path)
    store[path + attrs_key] = json_dumps(attrs)


def create_meta_store(reader: ReaderBase, tilesize: int) -> Dict[str, bytes]:
    """Creates a dict containing the zarr metadata for the multiscale openslide image."""
    store = dict()
    root_attrs = {
        "multiscales": [
            {
                "name": reader.file,
                "datasets": [
                    {"path": str(i)} for i in range(reader.properties.n_level)
                ],
                "version": "0.1",
            }
        ],
        "metadata": asdict(reader.properties),
    }
    init_group(store)
    init_attrs(store, root_attrs)
    for i, (x, y) in enumerate(reader.properties.level_shape):
        init_array(
            store,
            path=str(i),
            shape=normalize_shape((x, y, 3)),
            chunks=(tilesize, tilesize, 3),
            fill_value=0,
            dtype="|u1",
            compressor=None,
        )
        suffix = str(i) if i != 0 else ""
        init_attrs(
            store, {"_ARRAY_DIMENSIONS": [f"Y{suffix}", f"X{suffix}", "S"]}, path=str(i)
        )
    return store


def _parse_chunk_path(path: str):
    """Returns x,y chunk coords and pyramid level from string key"""
    level, ckey = path.split("/")
    y, x, _ = map(int, ckey.split("."))
    return x, y, int(level)


class ReaderStore(Store):
    """Wraps a Reader object as a multiscale Zarr Store.

    Parameters
    ----------
    reader: Reader
        The reader object
    tilesize: int
        Desired "chunk" size for zarr store (default: 256).
    """

    def __init__(self, reader: ReaderBase, tilesize: int = 256):
        self._reader = reader
        self._tilesize = tilesize
        self._store = create_meta_store(reader, tilesize)
        self._writeable = False
        self._erasable = False

    def __getitem__(self, key: str):
        if key in self._store:
            # key is for metadata
            return self._store[key]

        # key should now be a path to an array chunk
        # e.g '3/4.5.0' -> '<level>/<chunk_key>'
        try:
            x, y, level = _parse_chunk_path(key)
            location = self._ref_pos(x, y, level)
            size = (self._tilesize, self._tilesize)
            tile = self._reader.get_region(*location, *size, level=level)
        except ArgumentError as err:
            # Can occur if trying to read a closed slide
            raise err
        except Exception:
            # TODO: probably need better error handling.
            # If anything goes wrong, we just signal the chunk
            # is missing from the store.
            raise KeyError(key)
        return np.array(tile)

    def __eq__(self, other):
        return (
            isinstance(other, ReaderStore)
            and self._reader.name == other._reader.name
            and self._reader.file == other._reader.file
        )

    def __setitem__(self, key, val):
        raise PermissionError("ZarrStore is read-only")

    def __delitem__(self, key):
        raise PermissionError("ZarrStore is read-only")

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return sum(1 for _ in self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _ref_pos(self, x: int, y: int, level: int):
        level = self._reader.translate_level(level)
        dsample = self._reader.properties.level_downsample[level]
        xref = int(x * dsample * self._tilesize)
        yref = int(y * dsample * self._tilesize)
        return xref, yref

    def keys(self):
        return self._store.keys()

    def close(self):
        self._reader.detach_reader()

    # Retrieved from napari-lazy-openslide PR#16
    def __getstate__(self):
        return {"_path": self._reader.file, "_tilesize": self._tilesize}

    def __setstate__(self, newstate):
        path = newstate["_path"]
        tilesize = newstate["_tilesize"]
        self.__init__(path, tilesize)

    def rename(self, path: str, new_path: str):
        raise PermissionError(f'{type(self)} is not erasable, cannot call "rename"')

    def rmdir(self, path: str = "") -> None:
        raise PermissionError(f'{type(self)} is not erasable, cannot call "rmdir"')


@singledispatch
def create_reader_store(reader: ReaderBase, tilesize: int = 256) -> KVStore:
    """Creates a ReaderStore from a Reader object."""
    return KVStore(ReaderStore(reader, tilesize=tilesize))


@create_reader_store.register(TiffSlideReader)
def _(reader: TiffSlideReader, tilesize: int = 256) -> KVStore:
    return reader.reader.zarr_group.store


def to_datatree(
    reader,
    chunks=(3, 256, 256),
) -> DataTree:
    store = create_reader_store(reader)
    img_dataset = open_zarr(store, consolidated=False, mask_and_scale=False)
    images = {}
    for level, key in enumerate(sorted(list(img_dataset.keys()), key=int)):
        suffix = key if key != "0" else ""

        scale_image = DataArray(
            img_dataset[key].transpose("S", f"Y{suffix}", f"X{suffix}"),
            dims=("c", "y", "x"),
        ).chunk(chunks)

        scale_factor = reader.properties.level_downsample[level]

        if scale_factor == 1:
            transform = Identity()
        else:
            transform = Scale([scale_factor, scale_factor], axes=("y", "x"))

        scale_image = Image2DModel.parse(
            scale_image,
            transformations={"global": transform},
            c_coords=["r", "g", "b"],
        )
        scale_image.coords["y"] = scale_factor * scale_image.coords["y"]
        scale_image.coords["x"] = scale_factor * scale_image.coords["x"]

        images[f"scale{key}"] = Dataset({"image": scale_image})

    slide_image = DataTree.from_dict(images)
    slide_image.attrs = asdict(reader.properties)
    return slide_image



class DatasetAccessor(object):
    """
    Accessor for dataset objects.
    """

    def __init__(self, obj):
        self._obj = obj

    def tile_images(
        self,
        tile_key: str = "tiles",
        target_key: str = None,
        transform=None,
        target_transform=None,
        color_norm=None,
    ):
        """
        Create a TileImagesDataset from the current object.

        Parameters
        ----------
        tile_key : str, default: "tiles"
            The key of the tile table.
        target_key : str
            The key of the target table.
        transform: callable
            The transformation for the input tiles.
        target_transform: callable
            The transformation for the target.
        color_norm: str

        Returns
        -------
        TileImagesDataset

        """
        from ..dataset.image import TileImagesDataset

        return TileImagesDataset(
            self._obj,
            key=tile_key,
            target_key=target_key,
            transform=transform,
            target_transform=target_transform,
            color_norm=color_norm,
        )

    def tile_feature(
        self,
        feature_key: str,
        target_key: str = None,
        target_transform=None,
    ):
        """
        Create a TileFeatureDataset from the current object.

        Parameters
        ----------
        feature_key : str
            The key of the feature table.
        target_key : str
            The key of the target table.
        target_transform: callable
            The transformation for the target.

        Returns
        -------
        TileFeatureDataset

        """

        from ..dataset.feature import TileFeatureDataset

        return TileFeatureDataset(
            self._obj,
            feature_key=feature_key,
            target_key=target_key,
            target_transform=target_transform,
        )



"""Data structures and I/O functions for whole-slide images (WSIs)."""

__version__ = "0.6.1"

import wsidata.dataset as dataset
from ._model import WSIData, TileSpec
from .io import open_wsi, agg_wsi
from .accessors import (
    register_wsidata_accessor,
    FetchAccessor,
    IterAccessor,
    DatasetAccessor,
)
from .reader import get_reader, SlideProperties
from ._normalizer import ColorNormalizer


__all__ = [
    "open_wsi",
    "agg_wsi",
    "dataset",
    "WSIData",
    "TileSpec",
    "register_wsidata_accessor",
    "FetchAccessor",
    "IterAccessor",
    "DatasetAccessor",
    "get_reader",
    "SlideProperties",
    "ColorNormalizer",
]



# Modified from pandas accessor implementation
from __future__ import annotations

import warnings

from wsidata._utils import find_stack_level


class CachedAccessor:
    """
    Custom property-like object.

    A descriptor for caching accessors.

    Parameters
    ----------
    name : str
        Namespace that will be accessed under, e.g. ``df.foo``.
    accessor : cls
        Class with the extension methods.

    Notes
    -----
    For accessor, The class's __init__ method assumes that one of
    ``Series``, ``DataFrame`` or ``Index`` as the
    single argument ``data``.
    """

    def __init__(self, name: str, accessor) -> None:
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            # we're accessing the attribute of the class, i.e., Dataset.geo
            return self._accessor
        accessor_obj = self._accessor(obj)
        # Replace the property with the accessor object. Inspired by:
        # https://www.pydanny.com/cached-property.html
        # We need to use object.__setattr__ because we overwrite __setattr__ on
        # NDFrame
        object.__setattr__(obj, self._name, accessor_obj)
        return accessor_obj


def _register_accessor(name: str, cls):
    def decorator(accessor):
        if hasattr(cls, name):
            warnings.warn(
                f"registration of accessor {repr(accessor)} under name "
                f"{repr(name)} for type {repr(cls)} is overriding a preexisting "
                f"attribute with the same name.",
                UserWarning,
                stacklevel=find_stack_level(),
            )
        setattr(cls, name, CachedAccessor(name, accessor))
        # cls.accessors.add(name)
        return accessor

    return decorator


def register_wsidata_accessor(name: str):
    """Register a custom accessor on WSIData objects.

    Examples
    --------

    Create a custom accessor for WSIData objects,
    the init method of the accessor class should accept a single argument,
    which is the WSIData object.

    .. code-block:: python

        >>> @register_wsidata_accessor("my_accessor")
        ... class MyAccessor:
        ...    def __init__(self, obj):
        ...        self.obj = obj
        ...
        ...    def my_method(self):
        ...        return "Hello, world!"


    """
    from .._model import WSIData

    return _register_accessor(name, WSIData)



from .base import ReaderBase, SlideProperties
from .openslide import OpenSlideReader
from .tiffslide import TiffSlideReader
from .bioformats import BioFormatsReader
from .cucim import CuCIMReader
from .utils import get_reader, try_reader
from ._reader_datatree import to_datatree
